setwd("~/Documents/Studying/Lund_PhD/epistles/02_article_triplets/triplets_analysis")

library(rstan); require(stringr); require(plyr); require(ggplot2)
source("source_for_multiple_cores.R")

####### Helper functions ###########

stan_row_to_matrix <- function(stan_row, param_name) {
  stan_row <- stan_row[str_detect(names(stan_row), paste0("^", param_name, "\\["))]
  param_name_match <- str_match(names(stan_row), paste0("^", param_name,"\\[(\\d+),(\\d+)\\]"))
  row_i <- as.numeric(param_name_match[,2])
  col_i <- as.numeric(param_name_match[,3])
  mat <- matrix(nrow = max(row_i), ncol = max(col_i))
  mat[cbind(row_i, col_i)] <- stan_row
  mat
}

colVars <- function(a) {
  n <- dim(a)[[1]]; 
  c <- dim(a)[[2]]; 
  return(.colMeans(((a - matrix(.colMeans(a, n, c), nrow = n, ncol = c, byrow = TRUE)) ^ 2), n, c) * n / (n - 1))
}

waic <- function(stanfit){
  log_lik <- extract (stanfit, "log_lik")$log_lik
  dim(log_lik) <- if (length(dim(log_lik))==1) c(length(log_lik),1) else
    c(dim(log_lik)[1], prod(dim(log_lik)[2:length(dim(log_lik))]))
  S <- nrow(log_lik)
  n <- ncol(log_lik)
  lpd <- log(colMeans(exp(log_lik)))
  p_waic <- colVars(log_lik)
  elpd_waic <- lpd - p_waic
  waic <- -2*elpd_waic
  loo_weights_raw <- 1/exp(log_lik-max(log_lik))
  loo_weights_normalized <- loo_weights_raw/
    matrix(colMeans(loo_weights_raw),nrow=S,ncol=n,byrow=TRUE)
  loo_weights_regularized <- pmin (loo_weights_normalized, sqrt(S))
  elpd_loo <- log(colMeans(exp(log_lik)*loo_weights_regularized)/
                    colMeans(loo_weights_regularized))
  p_loo <- lpd - elpd_loo
  pointwise <- cbind(waic,lpd,p_waic,elpd_waic,p_loo,elpd_loo)
  total <- colSums(pointwise)
  se <- sqrt(n*colVars(pointwise))
  return(list(waic=total["waic"], elpd_waic=total["elpd_waic"],
              p_waic=total["p_waic"], elpd_loo=total["elpd_loo"], p_loo=total["p_loo"],
              pointwise=pointwise, total=total, se=se))
}



####### Model starts here ###############


model_string <- '
data {
int<lower=1> n_triplets;
int<lower=1> n_dims; 
int<lower=1> n_items; 
int<lower=1, upper=n_items> triplet[n_triplets, 3];
int<lower=1, upper=3> choice[n_triplets];
}

transformed data {
row_vector[n_dims] origin;
origin <- rep_row_vector(0.0, n_dims);
}

parameters {
matrix[n_items - 1, n_dims] pos_except_first;
#  vector<lower = 0>[n_dims] lambda;
}

transformed parameters {
matrix[n_items, n_dims] pos;
pos <- append_row(origin, pos_except_first);
}

model {
vector[3] neg_dist; 
for(i in 1:n_triplets) {
if(triplet[i, 2] == triplet[i, 3]) {
neg_dist[1] <- 0;
} else {
neg_dist[1] <- -distance(pos[triplet[i, 2]], pos[triplet[i, 3]]) ;
}

if(triplet[i, 1] == triplet[i, 3]) {
neg_dist[2] <- 0;
} else {
neg_dist[2] <- -distance(pos[triplet[i, 1]], pos[triplet[i, 3]]) ;
}

if(triplet[i, 1] == triplet[i, 2]) {
neg_dist[3] <- 0;
} else {
neg_dist[3] <- -distance(pos[triplet[i, 1]], pos[triplet[i, 2]]) ;
}

choice[i] ~ categorical_logit(neg_dist);
}

for(dim_i in 1:n_dims) {
for(item_i in 1:(n_items - 1) ) {
pos_except_first[item_i, dim_i] ~ normal(0, 5);
}
}
}

generated quantities {

vector[3] neg_dist; 
vector[n_triplets] log_lik;
for(i in 1:n_triplets) {
if(triplet[i, 2] == triplet[i, 3]) {
neg_dist[1] <- 0;
} else {
neg_dist[1] <- -distance(pos[triplet[i, 2]], pos[triplet[i, 3]]) ;
}

if(triplet[i, 1] == triplet[i, 3]) {
neg_dist[2] <- 0;
} else {
neg_dist[2] <- -distance(pos[triplet[i, 1]], pos[triplet[i, 3]]) ;
}

if(triplet[i, 1] == triplet[i, 2]) {
neg_dist[3] <- 0;
} else {
neg_dist[3] <- -distance(pos[triplet[i, 1]], pos[triplet[i, 2]]) ;
}

log_lik[i] <- categorical_logit_log(choice[i], neg_dist);
}

}'


mds_model <- stan_model(model_code = model_string, model_name = "bayes_mds")


####### Loading in the data ########


d <- read.csv("triplets_output_participants-with-min-10-trials_9551-lines.csv", stringsAsFactors=FALSE)
# d = d[d$lang=='en',]  # limit the analysis to one language group, if desired
d = d[,c("sound1", "sound2", "sound3", "response")] # cut the crap
names_info <- read.csv("triplets_132names-with-acoustics.csv", stringsAsFactors=FALSE) # get labels
names_info <- names_info[ order(names_info$file), ]

item_mat <- as.matrix(d[,c("sound1", "sound2", "sound3")])
pos_names <- sort(unique(item_mat[seq_along(item_mat)]))
choices <- as.numeric(factor(d$response))
triplets <- item_mat
mode(triplets) <- "factor"

sound_category <- names_info$emotion # sound_category=names_info$emotion or sound_category=names_info$pred.em
sound_sex <- names_info$sex
# sound_sex[is.na(sound_sex)] <- "unknown"

item_cat_mat <- item_mat
item_cat_mat[,1:3] <- names_info$emotion[ match(item_mat[,1:3], names_info$file) ]

pos_cat_names <- sort(unique(item_cat_mat[seq_along(item_cat_mat)]))
choices_cat <- as.numeric(factor(d$response))
triplets_cat <- item_cat_mat
mode(triplets_cat) <- "factor"
n_items = max(triplets)
n_cats = max(triplets_cat)

############ Trying different number of dimensions and calculating waic #######
# require (plyr)
if(FALSE) {
  waic_fit <- ldply(1:10,function(n_dims) {
    data_list = list(choice = choices, triplet = triplets, n_triplets = nrow(triplets),
                     n_dims = n_dims, n_items = n_items)
    n_chains = 1
    fit <- sampling(mds_model, data = data_list, pars = c("pos", "log_lik"), chains = n_chains, iter = 1000, thin = 1)
    cbind(n_dims , data.frame(rbind(waic(fit)$total)))
  })
  # saveRDS(waic_fit, "/home/allgoodguys/Documents/Studying/Lund_cognitive-science/00_master/experiments/triplets/analysis/output_stan/waic_fit_2cats-per-trial.rds")
} else {
  waic_fit <- readRDS("waic_fit.rds")
}
# waic_fit = readRDS ("/home/allgoodguys/Documents/Studying/Lund_cognitive-science/00_master/experiments/triplets/analysis/output_stan/waic_fit_2cats-per-trial.rds")
plot(waic_fit$n_dims, waic_fit$elpd_loo, type="b", xlab="Number of dimensions", ylab = 'WAIC')




############### Fitting the model with two dimensions  ###########################


n_dims = 3
data_list = list(choice = choices, triplet = triplets, n_triplets = nrow(triplets),
                 n_dims = n_dims, n_items = n_items)

n_chains = 3
if(FALSE) {
  fit <- sampling(mds_model, data = data_list, pars = c("pos"), chains = n_chains, iter = 1000, thin = 1)
  saveRDS(fit, "stan_mcmc_2d.rds")
} else {
  fit <- readRDS("stan_mcmc_2d.rds")
}
# use 'stan' instead of 'sampling' command to use multiple cores (one core = one thread)
fit <- stan(model_string, pedantic=T, data = data_list, pars = c("pos"), chains = n_chains, iter = 500, thin = 1)


fit_mat <- as.matrix(fit)
# saveRDS(fit_mat, "stan_mcmc_mat_2d.rds")

# rstan::traceplot(fit, pars = c("pos[3, 1]", "pos[3, 2]","pos[4, 2]", "pos[5, 2]", "pos[6, 2]" ,"lp__"))

fit_dist_mat <- aaply(fit_mat[sample(nrow(fit_mat), size = 1000, replace = FALSE),], 1, .drop = FALSE, function(stan_row) {  
  as.matrix(
    dist(stan_row_to_matrix(stan_row, "pos"))
  )
})

# plot(fit_dist_mat[,1,2], type="l")
# plot(fit_dist_mat[,10,20], type="l")

med_dist <- aaply(fit_dist_mat, c(2,3), median)
rownames(med_dist) <- pos_names
colnames(med_dist) <- pos_names
write.csv (med_dist, "/home/allgoodguys/Documents/Studying/Lund_cognitive-science/00_master/experiments/triplets/analysis/output_stan/med_dist_2D_en.csv")
write.csv (med_dist, "/home/allgoodguys/Downloads/med_dist_3D_en.csv")


saveRDS(fit_dist_mat, "mcmc_distance_matrix_2d.rds")
saveRDS(med_dist, "median_distance_matrix_2d.rds")


### Picking out the best samples with the highest LP and use to generate a 
### point estimate (called map as in Maximum A Posterior estimate).

best_pos_mcmc_draw <- stan_row_to_matrix(fit_mat[which.max(fit_mat[,"lp__"]), grep("^pos", colnames(fit_mat))], "pos")
init_list <- list(pos_except_first = best_pos_mcmc_draw[-1,])
map_fit <- optimizing(mds_model, data = data_list, init=init_list) # for 1D, delete init=init_list
map_mat <- stan_row_to_matrix(map_fit$par, "pos")
map_pos <- as.data.frame(map_mat)
map_pos <- rename(map_pos, list(V1 = "d1", V2 = "d2"))
map_pos$item_name <- pos_names

# for 3D
best_pos_mcmc_draw <- stan_row_to_matrix(fit_mat[which.max(fit_mat[,"lp__"]), grep("^pos", colnames(fit_mat))], "pos")
init_list <- list(pos_except_first = best_pos_mcmc_draw[-1,])
map_fit <- optimizing(mds_model, data = data_list, init=init_list)
map_mat <- stan_row_to_matrix(map_fit$par, "pos")
map_pos <- as.data.frame(map_mat)
map_pos <- rename(map_pos, list(V1 = "d1", V2 = "d2", V3 = "d3"))
map_pos$item_name <- pos_names

# saveRDS(map_pos, "map_pos_2d.rds")
write.csv (map_pos, "/home/allgoodguys/Documents/Studying/Lund_cognitive-science/00_master/experiments/triplets/analysis/output_stan/map_pos_2D_sv.csv")

qplot(map_mat[,1], map_mat[,2] , geom="text", label = sound_category, color = sound_category)
qplot(map_mat[,1], map_mat[,2]  ,color = sound_category, size=I(4))
qplot(map_mat[,1], map_mat[,2] , geom="text", label = names_info$emotion, ,color = names_info$emotion)
# for simulations:
sounds = data.frame (sound = 1:132, cat = c(rep("a",14),rep("b",14),rep("c",14),rep("d",15),rep("e",15),rep("f",15),rep("g",15),rep("h",15),rep("i",15)))
qplot(x=map_mat[,1], y=map_mat[,2] , geom="text", label = sounds$cat, color = sounds$cat)
plot (prcomp(dist(map_pos)))

# for 1D: qplot(x=map_mat[,1], y=1, geom="text", label = sounds$cat, color = sounds$cat)
# for 3D: 
# library (rgl)
# plot3d (map_mat[,1], map_mat[,2], map_mat[,3], type="n")
# text3d (map_mat[,1], map_mat[,2], map_mat[,3], text=sounds$cat, col=as.numeric (sounds$cat)) 

###############
# This hack plots the positions of the items by transforming
# the estimated distance matrix using cmdscale

fit <- cmdscale(med_dist,eig=TRUE, k=3) # k is the number of dim
# plot solution
x <- fit$points[,1]
y <- fit$points[,2]
# z <- fit$points[,3]

qplot(x, y , geom="text", label = sound_category, ,color = sound_category)
qplot(x, y , geom="text", label = names_info$emotion, ,color = names_info$emotion)
qplot(x, y , geom="text", label = names_info$pred.em, ,color = names_info$pred.em)
# plot3d (x,y,z, type="n", xlab="",ylab="",zlab=""); text3d (x,y,z, text=sounds$cat, col=as.numeric (sounds$cat)) 
########### Here using clustering on the distance matrix #######

rownames(med_dist) <- names_info$pred.em
colnames(med_dist) <- names_info$pred.em
med_dist_as_dist <- as.dist(med_dist)
fit <- hclust(med_dist_as_dist)
plot(fit) # display dendogram
groups <- cutree(fit, k=4) # cut tree into 3 clusters
# draw dendogram with red borders around the 3 clusters
rect.hclust(fit, k=4, border="red") 

##################################
##################################
##################################
## Basically doing the same thing as above but with categories


############ Trying different number of dimensions and calculating waic #######


waic_fit <- ldply(1:8,function(n_dims) {
  data_list = list(choice = choices_cat, triplet = triplets_cat, n_triplets = nrow(triplets_cat),
                   n_dims = n_dims, n_items = n_cats)
  n_chains = 1
  fit <- sampling(mds_model, data = data_list, pars = c("pos", "log_lik"), chains = n_chains, iter = 1000, thin = 1)
  cbind(n_dims , data.frame(rbind(waic(fit)$total)))
})

plot(waic_fit$n_dims, waic_fit$elpd_loo, type="b")

write.csv(waic_fit, "waic_cat_pred-em.csv")

############### Fitting the model with two dimensions  ###########################


n_dims = 2
data_list = list(choice = choices_cat, triplet = triplets_cat, n_triplets = nrow(triplets_cat),
                 n_dims = n_dims, n_items = n_cats)

n_chains = 3
fit <- sampling(mds_model, data = data_list, pars = c("pos"), chains = n_chains, iter = 5000, thin = 1)

# use 'stan' instead of 'sampling' command to use multiple cores (one core = one thread)
fit <- stan(model_string, pedantic=T, data = data_list, pars = c("pos"), chains = n_chains, iter = 1000, thin = 1)

fit_mat <- as.matrix(fit)

# saveRDS(fit, "cat_stan_mcmc_2d_true-em.rds")
# saveRDS(fit_mat, "cat_stan_mcmc_mat_2d_true-em.rds")

rstan::traceplot(fit, pars = c("pos[3, 1]", "pos[3, 2]","pos[4, 2]", "pos[5, 2]", "pos[6, 2]" ,"lp__"))

fit_dist_mat <- aaply(fit_mat[sample(nrow(fit_mat), size = 1000, replace = FALSE),], 1, .drop = FALSE, function(stan_row) {  
  as.matrix(
    dist(stan_row_to_matrix(stan_row, "pos"))
  )
})

plot(fit_dist_mat[,2,5], type="l")


med_dist <- aaply(fit_dist_mat, c(2,3), median)
rownames(med_dist) <- pos_cat_names
colnames(med_dist) <- pos_cat_names

saveRDS(fit_dist_mat, "cat_mcmc_distance_matrix_2d_true-em.rds")
saveRDS(med_dist, "cat_median_distance_matrix_2d_true-em.rds")

### Picking out the best samples with the highest LP and use to generate a 
### point estimate (called map as in Maximum A Posterior estimate).

best_pos_mcmc_draw <- stan_row_to_matrix(fit_mat[which.max(fit_mat[,"lp__"]), grep("^pos", colnames(fit_mat))], "pos")
init_list <- list(pos_except_first = best_pos_mcmc_draw[-1,])
map_fit <- optimizing(mds_model, data = data_list, init=init_list)
map_mat <- stan_row_to_matrix(map_fit$par, "pos")
map_pos <- as.data.frame(map_mat)
map_pos <- rename(map_pos, list(V1 = "d1", V2 = "d2"))
map_pos$item_name <- pos_cat_names

saveRDS(map_pos, "cat_map_pos_2d_true-em.rds")

qplot(map_mat[,1], map_mat[,2] , geom=c("text","point"), label = pos_cat_names)

###############
# This hack plots the positions of the categories by transforming
# the estimated distance matrix using cmdscale

fit <- cmdscale(med_dist,eig=TRUE, k=2) # k is the number of dim
# plot solution
x <- fit$points[,1]
y <- fit$points[,2]

qplot(x, y , geom="text", label = pos_cat_names)

########### Here using clustering on the distance matrix #######

rownames(med_dist) <- pos_cat_names
colnames(med_dist) <- pos_cat_names
med_dist_as_dist <- as.dist(med_dist)
fit <- hclust(med_dist_as_dist)
plot(fit) # display dendogram
groups <- cutree(fit, k=4) # cut tree into 3 clusters
# draw dendogram with red borders around the 3 clusters
rect.hclust(fit, k=4, border="red") 


fit_mat = readRDS("/home/allgoodguys/Documents/Studying/Lund_cognitive-science/00_master/experiments/triplets/analysis/rasmus/cat_stan_mcmc_mat_3d_real-em.rds")
